from __future__ import annotations

import logging
import warnings
from functools import partial
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch
from scipy.sparse import csr_matrix, vstack

from scvi import settings
from scvi._constants import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import (
    CategoricalJointObsField,
    CategoricalObsField,
    LayerField,
    NumericalJointObsField,
)
from scvi.model._utils import (
    _get_batch_code_from_category,
    scatac_raw_counts_properties,
)
from scvi.model.base import UnsupervisedTrainingMixin
from scvi.model.base._de_core import _de_core
from scvi.module import PEAKVAE
from scvi.train._callbacks import SaveBestState
from scvi.utils._docstrings import de_dsp, devices_dsp, setup_anndata_dsp

from .base import ArchesMixin, BaseModelClass, VAEMixin

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from typing import Literal

    from anndata import AnnData

logger = logging.getLogger(__name__)


class PEAKVI(ArchesMixin, VAEMixin, UnsupervisedTrainingMixin, BaseModelClass):
    """Peak Variational Inference for chromatin accessilibity analysis :cite:p:`Ashuach22`.

    Parameters
    ----------
    adata
        AnnData object that has been registered via :meth:`~scvi.model.PEAKVI.setup_anndata`.
    n_hidden
        Number of nodes per hidden layer. If `None`, defaults to square root
        of number of regions.
    n_latent
        Dimensionality of the latent space. If `None`, defaults to square root
        of `n_hidden`.
    n_layers_encoder
        Number of hidden layers used for encoder NN.
    n_layers_decoder
        Number of hidden layers used for decoder NN.
    dropout_rate
        Dropout rate for neural networks
    model_depth
        Model sequencing depth / library size (default: True)
    region_factors
        Include region-specific factors in the model (default: True)
    latent_distribution
        One of

        * ``'normal'`` - Normal distribution (Default)
        * ``'ln'`` - Logistic normal distribution (Normal(0, I) transformed by softmax)
    deeply_inject_covariates
        Whether to deeply inject covariates into all layers of the decoder. If False (default),
        covariates will only be included in the input layer.
    **model_kwargs
        Keyword args for :class:`~scvi.module.PEAKVAE`

    Examples
    --------
    >>> adata = anndata.read_h5ad(path_to_anndata)
    >>> scvi.model.PEAKVI.setup_anndata(adata, batch_key="batch")
    >>> vae = scvi.model.PEAKVI(adata)
    >>> vae.train()

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/atac/PeakVI`
    """

    _module_cls = PEAKVAE

    def __init__(
        self,
        adata: AnnData,
        n_hidden: int | None = None,
        n_latent: int | None = None,
        n_layers_encoder: int = 2,
        n_layers_decoder: int = 2,
        dropout_rate: float = 0.1,
        model_depth: bool = True,
        region_factors: bool = True,
        use_batch_norm: Literal["encoder", "decoder", "none", "both"] = "none",
        use_layer_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        latent_distribution: Literal["normal", "ln"] = "normal",
        deeply_inject_covariates: bool = False,
        encode_covariates: bool = False,
        **model_kwargs,
    ):
        super().__init__(adata)

        n_cats_per_cov = (
            self.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY).n_cats_per_key
            if REGISTRY_KEYS.CAT_COVS_KEY in self.adata_manager.data_registry
            else []
        )

        self.module = self._module_cls(
            n_input_regions=self.summary_stats.n_vars,
            n_batch=self.summary_stats.n_batch,
            n_hidden=n_hidden,
            n_latent=n_latent,
            n_layers_encoder=n_layers_encoder,
            n_layers_decoder=n_layers_decoder,
            n_continuous_cov=self.summary_stats.get("n_extra_continuous_covs", 0),
            n_cats_per_cov=n_cats_per_cov,
            dropout_rate=dropout_rate,
            model_depth=model_depth,
            region_factors=region_factors,
            use_batch_norm=use_batch_norm,
            use_layer_norm=use_layer_norm,
            latent_distribution=latent_distribution,
            deeply_inject_covariates=deeply_inject_covariates,
            encode_covariates=encode_covariates,
            **model_kwargs,
        )
        self._model_summary_string = (
            f"PeakVI Model with params: \nn_hidden: {self.module.n_hidden}, "
            f"n_latent: {self.module.n_latent}, n_layers_encoder: {n_layers_encoder}, "
            f"n_layers_decoder: {n_layers_decoder} , dropout_rate: {dropout_rate}, "
            f"latent_distribution: {latent_distribution}, "
            f"deep injection: {deeply_inject_covariates}, encode_covariates: {encode_covariates}"
        )
        self.n_latent = n_latent
        self.init_params_ = self._get_init_params(locals())

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int = 500,
        lr: float = 1e-4,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float | None = None,
        validation_size: float | None = None,
        shuffle_set_split: bool = True,
        batch_size: int = 128,
        weight_decay: float = 1e-3,
        eps: float = 1e-08,
        early_stopping: bool = True,
        early_stopping_patience: int = 50,
        save_best: bool = True,
        check_val_every_n_epoch: int | None = None,
        n_steps_kl_warmup: int | None = None,
        n_epochs_kl_warmup: int | None = 50,
        datasplitter_kwargs: dict | None = None,
        plan_kwargs: dict | None = None,
        **kwargs,
    ):
        """Trains the model using amortized variational inference.

        Parameters
        ----------
        max_epochs
            Number of passes through the dataset.
        lr
            Learning rate for optimization.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range [0.0, 1.0].
        validation_size
            Size of the test set. If `None`, defaults to 1 - `train_size`. If
            `train_size + validation_size < 1`, the remaining cells belong to a test set.
        shuffle_set_split
            Whether to shuffle indices before splitting. If `False`, the val, train, and test set
            are split in the sequential order of the data according to `validation_size` and
            `train_size` percentages.
        batch_size
            Minibatch size to use during training.
        weight_decay
            weight decay regularization term for optimization
        eps
            Optimizer eps
        early_stopping
            Whether to perform early stopping with respect to the validation set.
        early_stopping_patience
            How many epochs to wait for improvement before early stopping
        save_best
            ``DEPRECATED`` Save the best model state with respect to the validation loss (default),
            or use the final state in the training procedure
        check_val_every_n_epoch
            Check val every n train epochs. By default, val is not checked, unless `early_stopping`
            is `True`. If so, val is checked every epoch.
        n_steps_kl_warmup
            Number of training steps (minibatches) to scale weight on KL divergences from 0 to 1.
            Only activated when `n_epochs_kl_warmup` is set to None. If `None`, defaults
            to `floor(0.75 * adata.n_obs)`.
        n_epochs_kl_warmup
            Number of epochs to scale weight on KL divergences from 0 to 1.
            Overrides `n_steps_kl_warmup` when both are not `None`.
        datasplitter_kwargs
            Additional keyword arguments passed into :class:`~scvi.dataloaders.DataSplitter`.
        plan_kwargs
            Keyword args for :class:`~scvi.train.TrainingPlan`. Keyword arguments passed to
            `train()` will overwrite values present in `plan_kwargs`, when appropriate.
        **kwargs
            Other keyword args for :class:`~scvi.train.Trainer`.

        Notes
        -----
        ``save_best`` is deprecated in v1.2 and will be removed in v1.3. Please use
        ``enable_checkpointing`` instead.
        """
        update_dict = {
            "lr": lr,
            "weight_decay": weight_decay,
            "eps": eps,
            "n_epochs_kl_warmup": n_epochs_kl_warmup,
            "n_steps_kl_warmup": n_steps_kl_warmup,
            "optimizer": "AdamW",
        }
        if plan_kwargs is not None:
            plan_kwargs.update(update_dict)
        else:
            plan_kwargs = update_dict
        if save_best:
            warnings.warn(
                "`save_best` is deprecated in v1.2 and will be removed in v1.3. Please use "
                "`enable_checkpointing` instead. See "
                "https://github.com/scverse/scvi-tools/issues/2568 for more details.",
                DeprecationWarning,
                stacklevel=settings.warnings_stacklevel,
            )

            if "callbacks" not in kwargs.keys():
                kwargs["callbacks"] = []
            kwargs["callbacks"].append(SaveBestState(monitor="reconstruction_loss_validation"))

        super().train(
            max_epochs=max_epochs,
            train_size=train_size,
            accelerator=accelerator,
            devices=devices,
            validation_size=validation_size,
            shuffle_set_split=shuffle_set_split,
            early_stopping=early_stopping,
            early_stopping_monitor="reconstruction_loss_validation",
            early_stopping_patience=early_stopping_patience,
            datasplitter_kwargs=datasplitter_kwargs,
            plan_kwargs=plan_kwargs,
            check_val_every_n_epoch=check_val_every_n_epoch,
            batch_size=batch_size,
            **kwargs,
        )

    @torch.inference_mode()
    def get_library_size_factors(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] = None,
        batch_size: int = 128,
    ) -> dict[str, np.ndarray]:
        """Return library size factors.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.

        Returns
        -------
        Library size factor for expression and accessibility
        """
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        library_sizes = []
        for tensors in scdl:
            inference_inputs = self.module._get_inference_input(tensors)
            outputs = self.module.inference(**inference_inputs)
            library_sizes.append(outputs["d"].cpu())

        return torch.cat(library_sizes).numpy().squeeze()

    @torch.inference_mode()
    def get_region_factors(self):
        """Return region-specific factors."""
        if self.module.region_factors is None:
            raise RuntimeError("region factors were not included in this model")
        return torch.sigmoid(self.module.region_factors).cpu().numpy()

    @torch.inference_mode()
    def get_accessibility_estimates(
        self,
        adata: AnnData | None = None,
        indices: Sequence[int] = None,
        n_samples_overall: int | None = None,
        region_list: Sequence[str] | None = None,
        transform_batch: str | int | None = None,
        use_z_mean: bool = True,
        threshold: float | None = None,
        normalize_cells: bool = False,
        normalize_regions: bool = False,
        batch_size: int = 128,
        return_numpy: bool = False,
    ) -> pd.DataFrame | np.ndarray | csr_matrix:
        """Impute the full accessibility matrix.

        Returns a matrix of accessibility probabilities for each cell and genomic region in the
        input (for return matrix A, A[i,j] is the probability that region j is accessible in cell
        i).

        Parameters
        ----------
        adata
            AnnData object that has been registered with scvi. If `None`, defaults to the
            AnnData object used to initialize the model.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        n_samples_overall
            Number of samples to return in total
        region_list
            Return accessibility estimates for this subset of regions. if `None`, all regions are
            used. This can save memory when dealing with large datasets.
        transform_batch
            Batch to condition on.
            If transform_batch is:

            - None, then real observed batch is used
            - int, then batch transform_batch is used
        use_z_mean
            If True (default), use the distribution mean. Otherwise, sample from the distribution.
        threshold
            If provided, values below the threshold are replaced with 0 and a sparse matrix
            is returned instead. This is recommended for very large matrices. Must be between 0 and
            1.
        normalize_cells
            Whether to reintroduce library size factors to scale the normalized probabilities.
            This makes the estimates closer to the input, but removes the library size correction.
            False by default.
        normalize_regions
            Whether to reintroduce region factors to scale the normalized probabilities. This makes
            the estimates closer to the input, but removes the region-level bias correction. False
            by default.
        batch_size
            Minibatch size for data loading into model
        return_numpy
            If `True` and `threshold=None`, return :class:`~numpy.ndarray`. If `True` and
            `threshold` is given, return :class:`~scipy.sparse.csr_matrix`. If `False`, return
            :class:`~pandas.DataFrame`. DataFrame includes regions names as columns.
        """
        adata = self._validate_anndata(adata)
        adata_manager = self.get_anndata_manager(adata, required=True)
        if indices is None:
            indices = np.arange(adata.n_obs)
        if n_samples_overall is not None:
            indices = np.random.choice(indices, n_samples_overall)
        post = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)
        transform_batch = _get_batch_code_from_category(adata_manager, transform_batch)

        if region_list is None:
            region_mask = slice(None)
        else:
            all_regions = adata.var_names
            region_mask = [region in region_list for region in all_regions]

        if threshold is not None and (threshold < 0 or threshold > 1):
            raise ValueError("the provided threshold must be between 0 and 1")

        imputed = []
        for tensors in post:
            get_generative_input_kwargs = {"transform_batch": transform_batch[0]}
            generative_kwargs = {"use_z_mean": use_z_mean}
            inference_outputs, generative_outputs = self.module.forward(
                tensors=tensors,
                get_generative_input_kwargs=get_generative_input_kwargs,
                generative_kwargs=generative_kwargs,
                compute_loss=False,
            )
            p = generative_outputs["p"].cpu()

            if normalize_cells:
                p *= inference_outputs["d"].cpu()
            if normalize_regions:
                p *= torch.sigmoid(self.module.region_factors).cpu()
            if threshold:
                p[p < threshold] = 0
                p = csr_matrix(p.numpy())
            if region_list is not None:
                p = p[:, region_mask]
            imputed.append(p)

        if threshold:  # imputed is a list of csr_matrix objects
            imputed = vstack(imputed, format="csr")
        else:  # imputed is a list of tensors
            imputed = torch.cat(imputed).numpy()

        if return_numpy:
            return imputed
        elif threshold:
            return pd.DataFrame.sparse.from_spmatrix(
                imputed,
                index=adata.obs_names[indices],
                columns=adata.var_names[region_mask],
            )
        else:
            return pd.DataFrame(
                imputed,
                index=adata.obs_names[indices],
                columns=adata.var_names[region_mask],
            )

    @de_dsp.dedent
    def differential_accessibility(
        self,
        adata: AnnData | None = None,
        groupby: str | None = None,
        group1: Iterable[str] | None = None,
        group2: str | None = None,
        idx1: Sequence[int] | Sequence[bool] | str | None = None,
        idx2: Sequence[int] | Sequence[bool] | str | None = None,
        mode: Literal["vanilla", "change"] = "change",
        delta: float = 0.05,
        batch_size: int | None = None,
        all_stats: bool = True,
        batch_correction: bool = False,
        batchid1: Iterable[str] | None = None,
        batchid2: Iterable[str] | None = None,
        fdr_target: float = 0.05,
        silent: bool = False,
        two_sided: bool = True,
        **kwargs,
    ) -> pd.DataFrame:
        r"""\.

        A unified method for differential accessibility analysis.

        Implements `"vanilla"` DE :cite:p:`Lopez18`. and `"change"` mode DE :cite:p:`Boyeau19`.

        Parameters
        ----------
        %(de_adata)s
        %(de_groupby)s
        %(de_group1)s
        %(de_group2)s
        %(de_idx1)s
        %(de_idx2)s
        %(de_mode)s
        %(de_delta)s
        %(de_batch_size)s
        %(de_all_stats)s
        %(de_batch_correction)s
        %(de_batchid1)s
        %(de_batchid2)s
        %(de_fdr_target)s
        %(de_silent)s
        two_sided
            Whether to perform a two-sided test, or a one-sided test.
        **kwargs
            Keyword args for :meth:`scvi.model.base.DifferentialComputation.get_bayes_factors`

        Returns
        -------
        Differential accessibility DataFrame with the following columns:
        prob_da
            the probability of the region being differentially accessible
        is_da_fdr
            whether the region passes a multiple hypothesis correction procedure with the
            target_fdr threshold
        bayes_factor
            Bayes Factor indicating the level of significance of the analysis
        effect_size
            the effect size, computed as (accessibility in population 2) - (accessibility in
            population 1)
        emp_effect
            the empirical effect, based on observed detection rates instead of the estimated
            accessibility scores from the PeakVI model
        est_prob1
            the estimated probability of accessibility in population 1
        est_prob2
            the estimated probability of accessibility in population 2
        emp_prob1
            the empirical (observed) probability of accessibility in population 1
        emp_prob2
            the empirical (observed) probability of accessibility in population 2

        """
        adata = self._validate_anndata(adata)
        col_names = adata.var_names
        model_fn = partial(
            self.get_accessibility_estimates, use_z_mean=False, batch_size=batch_size
        )

        # TODO check if change_fn in kwargs and raise error if so
        def change_fn(a, b):
            return a - b

        if two_sided:

            def m1_domain_fn(samples):
                return np.abs(samples) >= delta

        else:

            def m1_domain_fn(samples):
                return samples >= delta

        result = _de_core(
            adata_manager=self.get_anndata_manager(adata, required=True),
            model_fn=model_fn,
            representation_fn=None,
            groupby=groupby,
            group1=group1,
            group2=group2,
            idx1=idx1,
            idx2=idx2,
            all_stats=all_stats,
            all_stats_fn=scatac_raw_counts_properties,
            col_names=col_names,
            mode=mode,
            batchid1=batchid1,
            batchid2=batchid2,
            delta=delta,
            batch_correction=batch_correction,
            fdr=fdr_target,
            change_fn=change_fn,
            m1_domain_fn=m1_domain_fn,
            silent=silent,
            **kwargs,
        )

        # manually change the results DataFrame to fit a PeakVI differential accessibility results
        result = pd.DataFrame(
            {
                "prob_da": result.proba_de,
                "is_da_fdr": result.loc[:, f"is_de_fdr_{fdr_target}"],
                "bayes_factor": result.bayes_factor,
                "effect_size": result.scale2 - result.scale1,
                "emp_effect": result.emp_mean2 - result.emp_mean1,
                "est_prob1": result.scale1,
                "est_prob2": result.scale2,
                "emp_prob1": result.emp_mean1,
                "emp_prob2": result.emp_mean2,
            },
        )
        return result

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        batch_key: str | None = None,
        labels_key: str | None = None,
        categorical_covariate_keys: list[str] | None = None,
        continuous_covariate_keys: list[str] | None = None,
        layer: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        %(param_batch_key)s
        %(param_labels_key)s
        %(param_cat_cov_keys)s
        %(param_cont_cov_keys)s
        %(param_layer)s
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
            CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
            CategoricalJointObsField(REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys),
            NumericalJointObsField(REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)



from __future__ import annotations

import json
import os
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING

from huggingface_hub import ModelCard, ModelCardData

from scvi.data import AnnDataManager
from scvi.data._utils import _is_minified
from scvi.model.base._save_load import _load_saved_files

from ._constants import _SCVI_HUB
from ._template import (
    condscvi_pretext,
    scanvi_pretext,
    scvi_pretext,
    stereoscope_pretext,
    template,
    totalvi_pretext,
)
from ._url import validate_url

if TYPE_CHECKING:
    import torch


@dataclass
class HubMetadata:
    """Encapsulates the required metadata for `scvi-tools` hub models.

    Parameters
    ----------
    scvi_version
        The version of `scvi-tools` that the model was trained with.
    anndata_version
        The version of anndata used during model training.
    model_cls_name
        The name of the model class.
    training_data_url
        Link to the training data used to train the model, if it is too large to be uploaded to the
        hub. This can be a cellxgene explorer session url. However it cannot be a self-hosted
        session -- it must be from the cellxgene portal (https://cellxgene.cziscience.com/).
    model_parent_module
        The parent module of the model class. Defaults to `scvi.model`. Change this if you are
        using a model class that is not in the `scvi.model` module, for example, if you are using a
        model class from a custom module.
    """

    scvi_version: str
    anndata_version: str
    model_cls_name: str
    training_data_url: str | None = None
    model_parent_module: str = _SCVI_HUB.DEFAULT_PARENT_MODULE

    @classmethod
    def from_dir(
        cls,
        local_dir: str,
        anndata_version: str,
        map_location: torch.device | str | dict | None = "cpu",
        **kwargs,
    ):
        """Create a `HubMetadata` object from a local directory.

        Parameters
        ----------
        local_dir
            The local directory containing the model files.
        anndata_version
            The version of anndata used during model training.
        map_location
            The device to map model tensors to, passed into :meth:`~torch.load`.
        kwargs
            Additional keyword arguments to pass to the HubMetadata initializer.
        """
        attr_dict, _, _, _ = _load_saved_files(
            local_dir, load_adata=False, map_location=map_location
        )
        scvi_version = attr_dict["registry_"]["scvi_version"]
        model_cls_name = attr_dict["registry_"]["model_name"]

        return cls(
            scvi_version,
            anndata_version,
            model_cls_name,
            **kwargs,
        )

    def save(self, save_path: str, overwrite: bool = False) -> None:
        """Save the metadata to a JSON file.

        Parameters
        ----------
        save_path
            The path to which to save the metadata as a JSON file.
        overwrite
            Whether to overwrite the file if it already exists.
        """
        if os.path.isfile(save_path) and not overwrite:
            raise FileExistsError(
                f"File already exists at {save_path}. To overwrite, pass `overwrite=True`."
            )
        with open(save_path, "w") as f:
            json.dump(asdict(self), f, indent=4)

    def __post_init__(self):
        if self.training_data_url is not None:
            validate_url(self.training_data_url, error_format=True)


@dataclass
class HubModelCardHelper:
    """A helper for creating a `ModelCard` for `scvi-tools` hub models.

    Parameters
    ----------
    license_info
        The license information for the model.
    model_cls_name
        The name of the model class.
    model_init_params
        The model initialization parameters.
    model_setup_anndata_args
        The arguments used to call ``setup_anndata`` during model training.
    model_summary_stats
        The model summary statistics.
    model_data_registry
        The model data registry.
    scvi_version
        The version of `scvi-tools` that the model was trained with.
    anndata_version
        The version of anndata used during model training.
    data_modalities
        The modalities of the training data.
    tissues
        The tissues of the training data.
    data_is_annotated
        Whether the training data is annotated.
    data_is_minified
        Whether the training data uploaded with the model has been minified.
    training_data_url
        Link to the training data used to train the model, if it is too large to be uploaded to the
        hub. This can be a cellxgene explorer session url. However it cannot be a self-hosted
        session -- it must be from the cellxgene portal (https://cellxgene.cziscience.com/).
    training_code_url
        Link to the code used to train the model.
    model_parent_module
        The parent module of the model class. Defaults to `scvi.model`. Change this if you are
        using a model class that is not in the `scvi.model` module, for example, if you are using a
        model class from a custom module.
    description
        A description of the model.
    references_
        A list of references for the model.
    metrics_report
        A dictionary containing the metrics report for the model.

    Notes
    -----
    It is not required to use this class to create a `ModelCard`. But this helps you do so in a way
    that is consistent with most other `scvi-tools` hub models. You can think of this as a
    template. The actual template string used can be found in ``scvi.template``. The resulting
    huggingface :class:`~huggingface_hub.ModelCard` can be accessed via the
    :meth:`~scvi.hub.HubModelCardHelper.model_card` property.
    """

    license_info: str
    model_cls_name: str
    model_init_params: dict
    model_setup_anndata_args: dict
    model_summary_stats: dict
    model_data_registry: dict
    scvi_version: str
    anndata_version: str
    data_modalities: list[str] = field(default_factory=list)
    tissues: list[str] = field(default_factory=list)
    data_is_annotated: bool | None = None
    data_is_minified: bool | None = None
    training_data_url: str | None = None
    training_code_url: str | None = None
    model_parent_module: str = _SCVI_HUB.DEFAULT_PARENT_MODULE
    description: str = _SCVI_HUB.DEFAULT_MISSING_FIELD
    references: str = _SCVI_HUB.DEFAULT_MISSING_FIELD
    metrics_report: str | None = None

    def __post_init__(self):
        self.model_card = self._to_model_card()

        if self.training_data_url is not None:
            validate_url(self.training_data_url, error_format=True)
        if self.training_code_url is not None:
            validate_url(self.training_code_url, error_format=True)

    @classmethod
    def from_dir(
        cls,
        local_dir: str,
        license_info: str,
        anndata_version: str,
        metrics_report: str | None = None,
        data_is_minified: bool | None = None,
        map_location: torch.device | str | dict | None = "cpu",
        **kwargs,
    ):
        """Create a `HubModelCardHelper` object from a local directory.

        Parameters
        ----------
        local_dir
            The local directory containing the model files.
        license_info
            The license information for the model.
        anndata_version
            The version of anndata used during model training.
        metrics_report
            Path to the json with stored metrics report.
        data_is_minified
            Whether the training data uploaded with the model has been minified.
        map_location
            The device to map model tensors to, passed into :meth:`~torch.load`.
        kwargs
            Additional keyword arguments to pass to the HubModelCardHelper initializer.
        """
        attr_dict, _, _, _ = _load_saved_files(
            local_dir, load_adata=False, map_location=map_location
        )
        model_init_params = attr_dict["init_params_"]
        registry = attr_dict["registry_"]
        model_cls_name = registry["model_name"]
        scvi_version = registry["scvi_version"]
        model_setup_anndata_args = registry["setup_args"]
        if metrics_report is None:
            if os.path.isfile(f"{local_dir}/metrics.json"):
                with open(f"{local_dir}/metrics.json") as f:
                    metrics_report = json.load(f)
            else:
                metrics_report = None
        else:
            with open(f"{local_dir}/metrics.json") as f:
                metrics_report = json.load(f)
        model_summary_stats = dict(AnnDataManager._get_summary_stats_from_registry(registry))
        model_data_registry = dict(AnnDataManager._get_data_registry_from_registry(registry))

        # get `is_minified` from the param if it is given, else from adata if it on disk, else set
        # it to None
        is_minified = data_is_minified
        if is_minified is None and os.path.isfile(f"{local_dir}/adata.h5ad"):
            is_minified = _is_minified(f"{local_dir}/adata.h5ad")

        return cls(
            license_info,
            model_cls_name,
            model_init_params,
            model_setup_anndata_args,
            model_summary_stats,
            model_data_registry,
            scvi_version,
            anndata_version,
            metrics_report=metrics_report,
            data_is_minified=is_minified,
            **kwargs,
        )

    def _to_model_card(self) -> ModelCard:
        # define tags
        tags = [
            "biology",
            "genomics",
            "single-cell",
            _SCVI_HUB.MODEL_CLS_NAME_TAG.format(self.model_cls_name),
            _SCVI_HUB.SCVI_VERSION_TAG.format(self.scvi_version),
            _SCVI_HUB.ANNDATA_VERSION_TAG.format(self.anndata_version),
        ]
        for m in self.data_modalities:
            tags.append(_SCVI_HUB.MODALITY_TAG.format(m))
        for t in self.tissues:
            tags.append(_SCVI_HUB.TISSUE_TAG.format(t))
        if self.data_is_annotated is not None:
            tags.append(_SCVI_HUB.ANNOTATED_TAG.format(self.data_is_annotated))

        # define the card data, which is the header
        card_data = ModelCardData(
            license=self.license_info,
            library_name=_SCVI_HUB.HF_LIBRARY_NAME,
            tags=tags,
        )

        # flatten the model_init_params into a single dict
        # for example {'kwargs': {'model_kwargs': {'foo': 'bar'}}, 'non_kwargs': {'n_hidden': 128,
        # 'n_latent': 10}}
        # becomes {'n_hidden': 128, 'n_latent': 10, 'foo': 'bar'}
        if "non_kwargs" in self.model_init_params.keys():
            non_kwargs = self.model_init_params["non_kwargs"]
            kwargs = self.model_init_params["kwargs"]
        else:
            non_kwargs = {
                k: v for k, v in self.model_init_params.items() if not isinstance(v, dict)
            }
            kwargs = {k: v for k, v in self.model_init_params.items() if isinstance(v, dict)}
        kwargs = {k: v for (i, j) in kwargs.items() for (k, v) in j.items()}
        # kwargs and non_kwargs keys should be disjoint but if not, we'll just use the original
        # model_init_params
        if len(set(kwargs.keys()).intersection(set(non_kwargs.keys()))) == 0:
            flattened_model_init_params = {**non_kwargs, **kwargs}
        else:
            flattened_model_init_params = self.model_init_params
        if self.model_cls_name == "SCVI":
            pretext = scvi_pretext
        elif self.model_cls_name == "SCANVI":
            pretext = scanvi_pretext
        elif self.model_cls_name == "TOTALVI":
            pretext = totalvi_pretext
        elif self.model_cls_name == "CondSCVI":
            pretext = condscvi_pretext
        elif self.model_cls_name == "RNAStereoscope":
            pretext = stereoscope_pretext
        else:
            raise ValueError(
                f"Model class name {self.model_cls_name} not recognized. Please provide a valid "
                "model class name for HuggingFace or create an issue to add this model to "
                "HuggingFace."
            )
        if self.metrics_report is not None:
            de_metrics = self.metrics_report.get("diff_exp", _SCVI_HUB.DEFAULT_NA_FIELD)
            cell_wise_cv = self.metrics_report.get("cell_wise_cv", _SCVI_HUB.DEFAULT_NA_FIELD)
            gene_wise_cv = self.metrics_report.get("gene_wise_cv", _SCVI_HUB.DEFAULT_NA_FIELD)
        else:
            de_metrics = _SCVI_HUB.DEFAULT_NA_FIELD
            cell_wise_cv = _SCVI_HUB.DEFAULT_NA_FIELD
            gene_wise_cv = _SCVI_HUB.DEFAULT_NA_FIELD

        # create the content from the template
        content = template.format(
            pretext=pretext,
            model_class_name=self.model_cls_name,
            card_data=card_data.to_yaml(),
            description=self.description,
            model_init_params=json.dumps(flattened_model_init_params, indent=4),
            model_setup_anndata_args=json.dumps(self.model_setup_anndata_args, indent=4),
            model_summary_stats=AnnDataManager._view_summary_stats(
                self.model_summary_stats, as_markdown=True
            ),
            model_data_registry=AnnDataManager._view_data_registry(
                self.model_data_registry, as_markdown=True
            ),
            model_parent_module=self.model_parent_module,
            data_is_minified=_SCVI_HUB.DEFAULT_MISSING_FIELD
            if self.data_is_minified is None
            else self.data_is_minified,
            training_data_url=self.training_data_url or _SCVI_HUB.DEFAULT_NA_FIELD,
            training_code_url=self.training_code_url or _SCVI_HUB.DEFAULT_NA_FIELD,
            references=self.references,
            de_metrics=de_metrics,
            cell_wise_cv=cell_wise_cv,
            gene_wise_cv=gene_wise_cv,
        )

        # finally create and return the actual card
        return ModelCard(content)



from typing import Literal

import torch
from anndata import AnnData

import scvi
from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager, fields
from scvi.train import TrainRunner


class TestSparseDataSplitter(scvi.dataloaders.DataSplitter):
    def __init__(self, *args, expected_sparse_layout: Literal["csr", "csc"] = None, **kwargs):
        if expected_sparse_layout == "csr":
            self.expected_sparse_layout = torch.sparse_csr
        elif expected_sparse_layout == "csc":
            self.expected_sparse_layout = torch.sparse_csc
        else:
            self.expected_sparse_layout = None

        super().__init__(*args, **kwargs)

    def on_after_batch_transfer(self, batch, dataloader_idx):
        X = batch.get(scvi.REGISTRY_KEYS.X_KEY)
        assert isinstance(X, torch.Tensor)
        assert X.layout is self.expected_sparse_layout

        batch = super().on_after_batch_transfer(batch, dataloader_idx)

        X = batch.get(scvi.REGISTRY_KEYS.X_KEY)
        assert isinstance(X, torch.Tensor)
        assert X.layout == torch.strided

        return batch


class TestSparseTrainingPlan(scvi.train.TrainingPlan):
    def training_step(self, batch, batch_idx):
        pass

    def validation_step(self, batch, batch_idx):
        pass


class TestSparseModule(scvi.module.base.BaseModuleClass):
    def __init__(self):
        super().__init__()
        self.layer = torch.nn.Linear(1, 1)

    def _get_inference_input(self):
        pass

    def _get_generative_input(self):
        pass

    def inference(self):
        pass

    def generative(self):
        pass

    def loss(self):
        pass


class TestSparseModel(scvi.model.base.BaseModelClass):
    def __init__(self, adata: AnnData):
        super().__init__(adata)
        self.module = TestSparseModule()

    @classmethod
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        batch_key: str | None = None,
    ):
        setup_method_args = cls._get_setup_method_args(**locals())
        anndata_fields = [
            fields.LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            fields.CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
        ]
        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata)
        cls.register_manager(adata_manager)

    def train(
        self,
        max_epochs: int = 1,
        accelerator: str = "auto",
        devices: int | list[int] | str = "auto",
        expected_sparse_layout: Literal["csr", "csc"] = None,
    ):
        data_splitter = TestSparseDataSplitter(
            self.adata_manager,
            expected_sparse_layout=expected_sparse_layout,
            load_sparse_tensor=True,
        )
        training_plan = TestSparseTrainingPlan(self.module)
        runner = TrainRunner(
            self,
            training_plan=training_plan,
            data_splitter=data_splitter,
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=devices,
        )
        return runner()



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from scvi import settings

from ._built_in_data._brain_large import _load_brainlarge_dataset
from ._built_in_data._cellxgene import _load_cellxgene_dataset
from ._built_in_data._cite_seq import (
    _load_pbmc_seurat_v4_cite_seq,
    _load_pbmcs_10x_cite_seq,
    _load_spleen_lymph_cite_seq,
)
from ._built_in_data._cortex import _load_cortex
from ._built_in_data._csv import _load_breast_cancer_dataset, _load_mouse_ob_dataset
from ._built_in_data._dataset_10x import _load_dataset_10x
from ._built_in_data._heartcellatlas import _load_heart_cell_atlas_subsampled
from ._built_in_data._loom import (
    _load_annotation_simulation,
    _load_frontalcortex_dropseq,
    _load_prefrontalcortex_starmap,
    _load_retina,
)
from ._built_in_data._pbmc import _load_pbmc_dataset, _load_purified_pbmc_dataset
from ._built_in_data._smfish import _load_smfish
from ._built_in_data._synthetic import _generate_synthetic

if TYPE_CHECKING:
    import anndata

    from scvi._types import AnnOrMuData


def pbmc_dataset(
    save_path: str = "data/",
    remove_extracted_data: bool = True,
) -> anndata.AnnData:
    """Loads pbmc dataset.

    We considered scRNA-seq data from two batches of peripheral blood mononuclear cells (PBMCs)
    from a healthy donor (4K PBMCs and 8K PBMCs). We derived quality control metrics using the
    cellrangerRkit R package (v. 1.1.0). Quality metrics were extracted from CellRanger throughout
    the molecule specific information file. After filtering, we extract 12,039 cells with 10,310
    sampled genes and get biologically meaningful clusters with the software Seurat. We then filter
    genes that we could not match with the bulk data used for differential expression to be left
    with g = 3346.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    remove_extracted_data
        If true, will remove the folder the data was extracted to

    Returns
    -------
    AnnData with batch info (``.obs['batch']``), label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.pbmc_dataset()
    """
    return _load_pbmc_dataset(
        save_path=save_path,
        remove_extracted_data=remove_extracted_data,
    )


def dataset_10x(
    dataset_name: str | None = None,
    filename: str | None = None,
    save_path: str = "data/10X",
    url: str = None,
    return_filtered: bool = True,
    remove_extracted_data: bool = False,
    **scanpy_read_10x_kwargs,
) -> anndata.AnnData:
    """Loads a file from `10x <http://cf.10xgenomics.com/>`_ website.

    Parameters
    ----------
    dataset_name
        Name of the dataset file. Has to be one of:
        "frozen_pbmc_donor_a", "frozen_pbmc_donor_b", "frozen_pbmc_donor_c",
        "fresh_68k_pbmc_donor_a", "cd14_monocytes", "b_cells", "cd34", "cd56_nk", "cd4_t_helper",
        "regulatory_t", "naive_t", "memory_t", "cytotoxic_t", "naive_cytotoxic", "pbmc8k",
        "pbmc4k", "t_3k", "t_4k", "neuron_9k", "pbmc_1k_protein_v3", "pbmc_10k_protein_v3",
        "malt_10k_protein_v3", "pbmc_1k_v2", "pbmc_1k_v3", "pbmc_10k_v3", "hgmm_1k_v2",
        "hgmm_1k_v3", "hgmm_5k_v3", "hgmm_10k_v3", "neuron_1k_v2", "neuron_1k_v3", "neuron_10k_v3",
        "heart_1k_v2", "heart_1k_v3", "heart_10k_v3", 5k_pbmc_protein_v3",
        "5k_pbmc_protein_v3_nextgem", 1M_neurons".
    filename
        manual override of the filename to write to.
    save_path
        Location to use when saving/loading the data.
    url
        manual override of the download remote location.
        Note that we already provide urls for most 10X datasets,
        which are automatically formed only using the ``dataset_name``.
    return_filtered
        Either `filtered` data or `raw` data.
    remove_extracted_data
        Whether to remove extracted archives in the case of `.tar.gz` downloads.
    **scanpy_read_10x_kwargs
        Kwargs for scanpy's read_10x function

    Returns
    -------
    adata initialized with 10x data

    Examples
    --------
    >>> import scvi
    >>> neuron = scvi.data.dataset10X("neuron_9k")
    """
    return _load_dataset_10x(
        dataset_name=dataset_name,
        filename=filename,
        save_path=save_path,
        url=url,
        return_filtered=return_filtered,
        remove_extracted_data=remove_extracted_data,
        **scanpy_read_10x_kwargs,
    )


def cellxgene(
    url: str,
    filename: str | None = None,
    save_path: str = "data/",
    return_path: bool = False,
) -> anndata.AnnData | str:
    """Loads a file from `cellxgene <https://cellxgene.cziscience.com/>`_ portal.

    Parameters
    ----------
    url
        URL to cellxgene session
    filename
        manual override of the filename to write to.
    save_path
        Location to use when saving/loading the data.
    return_path
        Whether to return the path to the downloaded file.

    Returns
    -------
    adata initialized with cellxgene data
    """
    # TODO: remove in 1.3.0
    warnings.warn(
        "The `cellxgene` function is deprecated and will be removed in scvi-tools 1.3. "
        "Please directly use the `cellxgene_census` package instead.",
        DeprecationWarning,
        stacklevel=settings.warnings_stacklevel,
    )

    return _load_cellxgene_dataset(
        url=url,
        filename=filename,
        save_path=save_path,
        return_path=return_path,
    )


def smfish(
    save_path: str = "data/",
    use_high_level_cluster: bool = True,
) -> anndata.AnnData:
    """Loads osmFISH data of mouse cortex cells from the Linarsson lab.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    use_high_level_cluster
        If True, use higher-level agglomerate clusters.
        The resulting cell types are "Astrocytes", "Endothelials", "Inhibitory",
        "Microglias", "Oligodendrocytes" and "Pyramidals".

    Returns
    -------
    AnnData with batch info (``.obs['batch']``), label info (``.obs['labels']``),
    spatial info (``.obs['x_coord']``, ``.obs['y_coord']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.smfish()
    """
    return _load_smfish(
        save_path=save_path,
        use_high_level_cluster=use_high_level_cluster,
    )


def purified_pbmc_dataset(
    save_path: str = "data/",
    subset_datasets: list[str] | None = None,
) -> anndata.AnnData:
    """Purified PBMC dataset.

    From: "Massively parallel digital transcriptional profiling of single cells".

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    subset_datasets
        index for subsetting the follwing list of datasets
        which are used to form the ``PurifiedPBMCDataset``:
        "cd4_t_helper", "regulatory_t", "naive_t", "memory_t", "cytotoxic_t", "naive_cytotoxic",
        "b_cells", "cd4_t_helper", "cd34", "cd56_nk", "cd14_monocytes".

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.purified_pbmc_dataset()
    """
    return _load_purified_pbmc_dataset(
        save_path=save_path,
        subset_datasets=subset_datasets,
    )


def prefrontalcortex_starmap(save_path: str = "data/") -> anndata.AnnData:
    """Loads a starMAP dataset of mouse pre-frontal cortex (Wang et al., 2018).

    3,704 cells and 166 genes.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``), label info (``.obs['labels']``),
    spatial info (``.obs['x_coord']``, ``.obs['y_coord']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.prefrontalcortex_starmap()

    """
    return _load_prefrontalcortex_starmap(save_path=save_path)


def frontalcortex_dropseq(save_path: str = "data/") -> anndata.AnnData:
    """Load the cells from the mouse frontal cortex (Saunders et al., 2018).

    Load the 71639 annotated cells located in the frontal cortex of adult mouses among the 690,000
    cells studied by (Saunders et al., 2018) using the Drop-seq method. We have a 71639*7611 gene
    expression matrix Among the 7611 genes, we offer the user to provide a list of genes to
    subsample from. If not provided, all genes are kept.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.frontalcortex_dropseq()
    """
    return _load_frontalcortex_dropseq(save_path=save_path)


def annotation_simulation(name: str, save_path: str = "data/") -> anndata.AnnData:
    """Simulated datasets for scANVI tutorials.

    Parameters
    ----------
    name
        One of "1", "2", or "3"
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.annontation_simulation("1")

    """
    return _load_annotation_simulation(name=name, save_path=save_path)


def retina(save_path: str = "data/") -> anndata.AnnData:
    """Loads retina dataset.

    The dataset of bipolar cells contains after their original pipeline for filtering 27,499 cells
    and13,166 genes coming from two batches. We use the cluster annotation from 15 cell-types from
    the author. We also extract their normalized data with Combat and use it for benchmarking.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> adata = retina()
    """
    return _load_retina(save_path=save_path)


def mouse_ob_dataset(save_path: str = "data/") -> anndata.AnnData:
    """Loads mouse ob dataset.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.mouse_ob_dataset()
    """
    return _load_mouse_ob_dataset(save_path=save_path)


def breast_cancer_dataset(save_path: str = "data/") -> anndata.AnnData:
    """Loads breast cancer dataset.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.breast_cancer_dataset()
    """
    return _load_breast_cancer_dataset(save_path=save_path)


def pbmcs_10x_cite_seq(
    save_path: str = "data/",
    protein_join: str = "inner",
) -> anndata.AnnData:
    """Filtered PBMCs from 10x Genomics profiled with RNA and protein.

    Datasets were filtered for doublets and other outliers as in
    https://github.com/YosefLab/totalVI_reproducibility/blob/master/data/data_filtering_scripts/pbmc_10k/pbmc_10k.py

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    protein_join
        Whether to take an inner join or outer join of proteins

    Returns
    -------
    AnnData with batch info (``.obs['batch']``),
    and protein expression (``.obsm["protein_expression"]``)

    Missing protein values are zero, when ``protein_join == "outer`` and are identified during
    ``AnnData`` setup.

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.pbmcs_10x_cite_seq()
    """
    return _load_pbmcs_10x_cite_seq(
        save_path=save_path,
        protein_join=protein_join,
    )


def spleen_lymph_cite_seq(
    save_path: str = "data/",
    protein_join: str = "inner",
    remove_outliers: bool = True,
) -> anndata.AnnData:
    """Immune cells from the murine spleen and lymph nodes :cite:p:`GayosoSteier21`.

    This dataset was used throughout the totalVI manuscript, and named SLN-all.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    protein_join
        Whether to take an inner join or outer join of proteins
    remove_outliers
        Whether to remove clusters annotated as doublet or low quality

    Returns
    -------
    AnnData with batch info (``.obs['batch']``), label info (``.obs['cell_types']``),
    protein expression (``.obsm["protein_expression"]``), and tissue (``.obs['tissue']``).

    Missing protein values are zero, when ``protein_join == "outer`` and are identified during
    ``AnnData`` setup.

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.spleen_lymph_cite_seq()
    """
    return _load_spleen_lymph_cite_seq(
        save_path=save_path,
        protein_join=protein_join,
        remove_outliers=remove_outliers,
    )


def brainlarge_dataset(
    save_path: str = "data/",
    sample_size_gene_var: int = 10000,
    max_cells_to_keep: int | None = None,
    n_genes_to_keep: int = 720,
    loading_batch_size: int = 100000,
) -> anndata.AnnData:
    """Loads brain-large dataset.

    This dataset contains 1.3 million brain cells from
    `10x Genomics <https://support.10xgenomics.com/single-cell-gene-expression/datasets>`_.
    We randomly shuffle the data to get a 1M subset of cells and order genes by variance to retain
    first 10,000 and then 720 sampled variable genes. This dataset is then sampled multiple times
    in cells for the runtime and goodness-of-fit analysis. We report imputation scores on the 10k
    cells and 720 genes samples only.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    sample_size_gene_var
        Number of cells to use to estimate gene variances.
    max_cells_to_keep
        Maximum number of cells to keep.
    n_genes_to_keep
        Number of genes to keep, ordered by decreasing variance.
    loading_batch_size
        Number of cells to use for each chunk loaded.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.brainlarge_dataset()
    """
    return _load_brainlarge_dataset(
        save_path=save_path,
        sample_size_gene_var=sample_size_gene_var,
        max_cells_to_keep=max_cells_to_keep,
        n_genes_to_keep=n_genes_to_keep,
        loading_batch_size=loading_batch_size,
    )


def cortex(save_path: str = "data/") -> anndata.AnnData:
    """Loads cortex dataset.

    The
    `Mouse Cortex Cells dataset <https://storage.googleapis.com/linnarsson-lab-www-blobs/blobs/cortex/expression_mRNA_17-Aug-2014.txt>`_
    contains 3005 mouse cortex cells and gold-standard labels for seven distinct cell types. Each
    cell type corresponds to a cluster to recover.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.

    Returns
    -------
    AnnData with batch info (``.obs['batch']``) and label info (``.obs['labels']``)

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.cortex()
    """
    return _load_cortex(save_path)


def synthetic_iid(
    batch_size: int = 200,
    n_genes: int = 100,
    n_proteins: int = 100,
    n_regions: int = 100,
    n_batches: int = 2,
    n_labels: int = 3,
    dropout_ratio: float = 0.7,
    sparse_format: str | None = None,
    generate_coordinates: bool = False,
    return_mudata: bool = False,
    **kwargs,
) -> AnnOrMuData:
    """Synthetic multimodal dataset.

    RNA and accessibility data are generated from a zero-inflated negative binomial,
    while protein data is generated from a negative binomial distribution. This dataset
    is just for testing purposes and not meant for modeling or research. Each value is
    independently and identically distributed.

    Parameters
    ----------
    batch_size
        The number of cells per batch such that the total number of cells in the data is
        `batch_size * n_batches`.
    n_genes
        The number of genes to generate.
    n_proteins
        The number of proteins to generate.
    n_regions
        The number of accessibility regions to generate.
    n_batches
        The number of batches to generate.
    n_labels
        The number of cell type labels, distributed uniformly across batches.
    sparse
        Whether to store ZINB generated data as a :class:`scipy.sparse.csr_matrix`.
    dropout_ratio
        The expected percentage of zeros artificially added into the data for RNA
        and accessibility data.
    sparse_format
        Whether to store RNA, accessibility, and protein data as sparse arrays. One of
        the following:

        * `None`: Store as a dense :class:`numpy.ndarray`.
        * `"csr_matrix"`: Store as a :class:`scipy.sparse.csr_matrix`.
        * `"csc_matrix"`: Store as a :class:`scipy.sparse.csc_matrix`.
    generate_coordinates
        Whether to generate spatial coordinates for the cells.
    return_mudata
        Returns a :class:`~mudata.MuData` if `True`, else :class:`~anndata.AnnData`.

    Returns
    -------
    :class:`~anndata.AnnData` (if `return_mudata=False`) with the following fields:

    * `.obs["batch"]`: Categorical batch labels in the format `batch_{i}`.
    * `.obs["labels"]`: Categorical cell type labels in the format `label_{i}`.
    * `.obsm["protein_expression"]`: Protein expression matrix.
    * `.uns["protein_names"]`: Array of protein names.
    * `.obsm["accessibility"]`: Accessibility expression matrix.
    * `.obsm["coordinates"]`: Spatial coordinates for the cells if ``generate_coordinates`` is
      ``True``.

    :class:`~mudata.MuData` (if `return_mudata=True`) with the following fields:

    * `.obs["batch"]`: Categorical batch labels in the format `batch_{i}`.
    * `.obs["labels"]`: Categorical cell type labels in the format `label_{i}`.
    * `.mod["rna"]`: RNA expression data.
    * `.mod["protein_expression"]`: Protein expression data.
    * `.mod["accessibility"]`: Accessibility expression data.
    * `.obsm["coordinates"]`: Spatial coordinates for the cells if ``generate_coordinates`` is
      ``True``.

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.synthetic_iid()
    """
    if n_batches < 1:
        raise ValueError("`n_batches` must be greater than 0")
    if n_genes < 1:
        raise ValueError("`n_genes` must be greater than 0")

    return _generate_synthetic(
        batch_size=batch_size,
        n_genes=n_genes,
        n_proteins=n_proteins,
        n_regions=n_regions,
        n_batches=n_batches,
        n_labels=n_labels,
        dropout_ratio=dropout_ratio,
        sparse_format=sparse_format,
        generate_coordinates=generate_coordinates,
        return_mudata=return_mudata,
        **kwargs,
    )


def heart_cell_atlas_subsampled(
    save_path: str = "data/",
    remove_nuisance_clusters: bool = True,
) -> anndata.AnnData:
    """Combined single cell and single nuclei RNA-Seq data of 485K cardiac cells with annotations.

    Dataset was filtered down randomly to 20k cells using :meth:`~scanpy.pp.subsample`. The
    original data can be downloaded from https://www.heartcellatlas.org/#DataSources.

    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    remove_nuisance_clusters
        Remove doublets and unsassigned cells

    Returns
    -------
    AnnData

    Notes
    -----
    The data were filtered using the following sequence::

        >>> adata = anndata.read_h5ad(path_to_anndata)
        >>> bdata = sc.pp.subsample(adata, n_obs=20000, copy=True)
        >>> sc.pp.filter_genes(bdata, min_counts=3)
        >>> bdata.write_h5ad(path, compression="gzip")

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.heart_cell_atlas_subsampled()
    """
    return _load_heart_cell_atlas_subsampled(
        save_path=save_path,
        remove_nuisance_clusters=remove_nuisance_clusters,
    )


def pbmc_seurat_v4_cite_seq(
    save_path: str = "data/",
    apply_filters: bool = True,
    aggregate_proteins: bool = True,
    mask_protein_batches: int = 0,
) -> anndata.AnnData:
    """Dataset of PBMCs measured with CITE-seq (161764 cells).

    This dataset was first presented in the Seurat v4 paper:

    https://doi.org/10.1016/j.cell.2021.04.048

    It contains 8 volunteers in an HIV vaccine trial measured
    at 3 time points; thus, there are 24 batches in this dataset.


    Parameters
    ----------
    save_path
        Location to use when saving/loading the data.
    apply_filters
        Apply filters at cell and protein level. At the cell level,
        this filters on protein library size, number proteins detected,
        percent mito, and removes cells labeled as doublets.
    aggregate_proteins
        Antibodies targeting the same surface protein are added together,
        and isotype controls are removed. See the source code for full details.
    mask_protein_batches
        Set proteins in this many batches to be all zero (considered missing
        for :class:`~scvi.model.TOTALVI`.). This improves transfer learning
        with this dataset.

    Returns
    -------
    AnnData

    Notes
    -----
    This is not the same exact dataset as can be downloaded from:

    https://satijalab.org/seurat/articles/multimodal_reference_mapping.html

    This is due to the fact that the object linked in the tutorial above does
    not contain the actual UMI count data for RNA. UMI counts had to be separately
    downloaded from GEO (GSE164378). The counts in that object are an output of the
    scTransform method and should not be treated like UMI counts.

    Examples
    --------
    >>> import scvi
    >>> adata = scvi.data.pbmc_seurat_v4_cite_seq()
    """
    return _load_pbmc_seurat_v4_cite_seq(
        save_path=save_path,
        apply_filters=apply_filters,
        aggregate_proteins=aggregate_proteins,
        mask_protein_batches=mask_protein_batches,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest

import scvi
from scvi.data import synthetic_iid
from scvi.data._constants import _ADATA_MINIFY_TYPE_UNS_KEY, ADATA_MINIFY_TYPE
from scvi.data._utils import _is_minified
from scvi.model import SCANVI, SCVI
from scvi.model.base import BaseMinifiedModeModelClass

if TYPE_CHECKING:
    import numpy.typing as npt
    from anndata import AnnData


def prep_model(
    cls: BaseMinifiedModeModelClass = SCVI,
    layer: str | None = None,
    use_size_factor: bool = False,
    n_latent: int = 5,
) -> tuple[BaseMinifiedModeModelClass, AnnData, npt.NDArray, AnnData]:
    adata = synthetic_iid()
    counts = adata.X
    if use_size_factor:
        adata.obs["size_factor"] = np.random.randint(1, 5, size=(adata.shape[0],))
    if layer is not None:
        adata.layers[layer] = adata.X.copy()
        adata.X = np.zeros_like(adata.X)

    adata_before_setup = adata.copy()

    setup_kwargs = {
        "layer": layer,
        "batch_key": "batch",
        "labels_key": "labels",
        "size_factor_key": "size_factor" if use_size_factor else None,
    }
    if cls == SCANVI:
        setup_kwargs["unlabeled_category"] = "unknown"
    cls.setup_anndata(
        adata,
        **setup_kwargs,
    )

    model = cls(adata, n_latent=n_latent)
    model.train(1, check_val_every_n_epoch=1, train_size=0.5)

    lib_size = np.squeeze(np.asarray(counts.sum(axis=-1)))

    return model, adata, lib_size, adata_before_setup


def assert_approx_equal(a, b):
    # Allclose because on GPU, the values are not exactly the same
    # as some values are moved to cpu during data minification
    np.testing.assert_allclose(a, b, rtol=3e-1, atol=5e-1)


def run_test_for_model_with_minified_adata(
    cls: BaseMinifiedModeModelClass = SCVI,
    n_samples: int = 1,
    give_mean: bool = False,
    layer: str = None,
    use_size_factor: bool = False,
):
    model, adata, adata_lib_size, _ = prep_model(cls, layer, use_size_factor)

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv
    scvi.settings.seed = 1
    params_orig = model.get_likelihood_parameters(n_samples=n_samples, give_mean=give_mean)
    adata_orig = adata.copy()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR
    assert model.adata_manager.registry is model.registry_

    assert not _is_minified(adata)
    assert adata is not model.adata

    orig_obs_df = adata_orig.obs
    orig_obs_df[BaseMinifiedModeModelClass._OBSERVED_LIB_SIZE_KEY] = adata_lib_size
    assert model.adata.obs.equals(orig_obs_df)
    assert model.adata.var_names.equals(adata_orig.var_names)
    assert model.adata.var.equals(adata_orig.var)

    scvi.settings.seed = 1
    keys = ["mean", "dispersions", "dropout"]
    if n_samples == 1:
        params_latent = model.get_likelihood_parameters(n_samples=n_samples, give_mean=give_mean)
    else:
        # do this so that we generate the same sequence of random numbers in the
        # minified and non-minified cases (purely to get the tests to pass). this is
        # because in the non-minified case we sample once more (in the call to z_encoder
        # during inference)
        params_latent = model.get_likelihood_parameters(n_samples=n_samples, give_mean=give_mean)
    for k in keys:
        assert params_latent[k].shape == params_orig[k].shape

    for k in keys:
        assert_approx_equal(params_latent[k], params_orig[k])


def test_scvi_with_minified_adata_one_sample():
    run_test_for_model_with_minified_adata()
    run_test_for_model_with_minified_adata(layer="data_layer", use_size_factor=True)


def test_scvi_with_minified_adata_one_sample_with_layer():
    run_test_for_model_with_minified_adata(layer="data_layer")
    run_test_for_model_with_minified_adata(layer="data_layer", use_size_factor=True)


def test_scvi_with_minified_adata_n_samples():
    run_test_for_model_with_minified_adata(n_samples=400, give_mean=True)
    run_test_for_model_with_minified_adata(n_samples=400, give_mean=True, use_size_factor=True)


def test_scanvi_with_minified_adata_one_sample():
    run_test_for_model_with_minified_adata(SCANVI)
    run_test_for_model_with_minified_adata(SCANVI, use_size_factor=True)


def test_scanvi_with_minified_adata_one_sample_with_layer():
    run_test_for_model_with_minified_adata(SCANVI, layer="data_layer")
    run_test_for_model_with_minified_adata(SCANVI, layer="data_layer", use_size_factor=True)


def test_scanvi_with_minified_adata_n_samples():
    run_test_for_model_with_minified_adata(SCANVI, n_samples=400, give_mean=True)
    run_test_for_model_with_minified_adata(
        SCANVI, n_samples=400, give_mean=True, use_size_factor=True
    )


def test_scanvi_from_scvi(save_path):
    model, adata, _, adata_before_setup = prep_model(SCVI)
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    adata.obsm["X_latent_qzm"] = qzm
    adata.obsm["X_latent_qzv"] = qzv
    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    # for from_scvi_model to succeed:
    # 1. scvi_model.adata must not be minified, and
    # 2. adata (if different from scvi_model.adata) must not be minified

    with pytest.raises(ValueError) as e:
        scvi.model.SCANVI.from_scvi_model(model, "label_0")

    msg = (
        "We cannot use the given scVI model to initialize scANVI because it has minified adata. "
        "Keep counts when minifying model using minified_data_type="
        "'latent_posterior_parameters_with_counts'."
    )
    assert str(e.value) == msg

    # let's load scvi_model with a non-minified adata
    model.save(save_path, overwrite=True)
    loaded_model = SCVI.load(save_path, adata=adata_before_setup)

    adata2 = synthetic_iid()
    # just add this to pretend the data is minified
    adata2.uns[_ADATA_MINIFY_TYPE_UNS_KEY] = ADATA_MINIFY_TYPE.LATENT_POSTERIOR
    with pytest.raises(ValueError) as e:
        scvi.model.SCANVI.from_scvi_model(loaded_model, "label_0", adata=adata2)
    assert str(e.value) == "Please provide a non-minified `adata` to initialize scANVI."

    scanvi_model = scvi.model.SCANVI.from_scvi_model(loaded_model, "label_0")
    scanvi_model.train(1)


def test_scvi_with_minified_adata_get_normalized_expression():
    model, adata, _, _ = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    exprs_orig = model.get_normalized_expression()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    exprs_new = model.get_normalized_expression()
    assert exprs_new.shape == adata.shape

    np.testing.assert_allclose(exprs_new, exprs_orig, rtol=3e-1, atol=5e-1)


def test_scvi_with_minified_adata_get_normalized_expression_non_default_gene_list():
    model, adata, _, _ = prep_model()

    # non-default gene list and n_samples > 1
    gl = adata.var_names[:5].to_list()
    n_samples = 10

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    exprs_orig = model.get_normalized_expression(
        gene_list=gl, n_samples=n_samples, library_size="latent"
    )

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    # do this so that we generate the same sequence of random numbers in the
    # minified and non-minified cases (purely to get the tests to pass). this is
    # because in the non-minified case we sample once more (in the call to z_encoder
    # during inference)
    exprs_new = model.get_normalized_expression(
        gene_list=gl, n_samples=n_samples + 1, return_mean=False, library_size="latent"
    )
    exprs_new = exprs_new[1:].mean(0)

    assert exprs_new.shape == (adata.shape[0], 5)
    np.testing.assert_allclose(exprs_new, exprs_orig, rtol=3e-1, atol=5e-1)


def test_validate_unsupported_if_minified():
    model, _, _, _ = prep_model()

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    common_err_msg = "The {} function currently does not support minified data."

    with pytest.raises(ValueError) as e:
        model.get_elbo()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_elbo")

    with pytest.raises(ValueError) as e:
        model.get_reconstruction_error()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_reconstruction_error")

    with pytest.raises(ValueError) as e:
        model.get_marginal_ll()
    assert str(e.value) == common_err_msg.format("VAEMixin.get_marginal_ll")

    with pytest.raises(ValueError) as e:
        model.get_latent_library_size()
    assert str(e.value) == common_err_msg.format("RNASeqMixin.get_latent_library_size")

    with pytest.raises(ValueError) as e:
        model.train()
    assert str(e.value) == common_err_msg.format("VAE.loss")


def test_validate_supported_if_minified_keep_count():
    model, _, _, _ = prep_model()
    model2, _, _, _ = prep_model()

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.minify_adata(minified_data_type="latent_posterior_parameters_with_counts")
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR_WITH_COUNTS
    assert model2.minified_data_type is None

    assert np.allclose(model2.get_elbo().cpu(), model.get_elbo().cpu(), rtol=5e-2)
    assert np.allclose(
        model2.get_reconstruction_error()["reconstruction_loss"].cpu(),
        model.get_reconstruction_error()["reconstruction_loss"].cpu(),
        rtol=5e-2,
    )
    assert np.allclose(model2.get_marginal_ll(), model.get_marginal_ll(), rtol=5e-2)

    model.train(1, check_val_every_n_epoch=1, train_size=0.5)
    model.train(
        1, check_val_every_n_epoch=1, train_size=0.5, plan_kwargs={"update_only_decoder": True}
    )
    scanvi_model = scvi.model.SCANVI.from_scvi_model(
        model, labels_key="labels", unlabeled_category="unknown"
    )
    scanvi_model.train()
    scanvi_model.train(
        1, check_val_every_n_epoch=1, train_size=0.5, plan_kwargs={"update_only_decoder": True}
    )


def test_scvi_with_minified_adata_save_then_load(save_path):
    # create a model and minify its adata, then save it and its adata.
    # Load it back up using the same (minified) adata. Validate that the
    # loaded model has the minified_data_type attribute set as expected.
    model, adata, _, _ = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    params_orig = model.get_likelihood_parameters()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    model.save(save_path, overwrite=True, save_anndata=True)
    # load saved model with saved (minified) adata
    loaded_model = SCVI.load(save_path)

    assert loaded_model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    params_latent = loaded_model.get_likelihood_parameters()
    assert params_latent["mean"].shape == adata.shape
    np.testing.assert_allclose(params_latent["mean"], params_orig["mean"], rtol=3e-1, atol=5e-1)


def test_scvi_with_minified_adata_save_then_load_with_non_minified_adata(save_path):
    # create a model and minify its adata, then save it and its adata.
    # Load it back up using a non-minified adata. Validate that the
    # loaded model does not has the minified_data_type attribute set.
    model, adata, _, adata_before_setup = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    params_orig = model.get_likelihood_parameters()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    model.save(save_path, overwrite=True, save_anndata=True)
    # load saved model with a non-minified adata
    loaded_model = SCVI.load(save_path, adata=adata_before_setup)

    assert loaded_model.minified_data_type is None

    scvi.settings.seed = 1
    params_new = loaded_model.get_likelihood_parameters()
    assert params_new["mean"].shape == adata.shape
    np.testing.assert_array_equal(params_new["mean"], params_orig["mean"])


def test_scvi_save_then_load_with_minified_adata(save_path):
    # create a model, then save it and its adata (non-minified).
    # Load it back up using a minified adata. Validate that this
    # fails, as expected because we don't have a way to validate
    # whether the minified-adata was set up correctly
    model, _, _, _ = prep_model()

    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    model.save(save_path, overwrite=True, save_anndata=True)

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    # loading this model with a minified adata is not allowed because
    # we don't have a way to validate whether the minified-adata was
    # set up correctly
    with pytest.raises(KeyError):
        SCVI.load(save_path, adata=model.adata)


def test_scvi_with_minified_adata_get_latent_representation():
    model, _, _, _ = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    latent_repr_orig = model.get_latent_representation()

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    latent_repr_new = model.get_latent_representation()

    np.testing.assert_array_equal(latent_repr_new, latent_repr_orig)


def test_scvi_with_minified_adata_posterior_predictive_sample():
    model, _, _, _ = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    sample_orig = model.posterior_predictive_sample(
        indices=[1, 2, 3], gene_list=["gene_1", "gene_2"]
    )

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    sample_new = model.posterior_predictive_sample(
        indices=[1, 2, 3], gene_list=["gene_1", "gene_2"]
    )
    assert sample_new.shape == (3, 2)

    np.testing.assert_array_equal(sample_new.todense(), sample_orig.todense())


def test_scvi_with_minified_adata_get_feature_correlation_matrix():
    model, _, _, _ = prep_model()

    scvi.settings.seed = 1
    qzm, qzv = model.get_latent_representation(give_mean=False, return_dist=True)
    model.adata.obsm["X_latent_qzm"] = qzm
    model.adata.obsm["X_latent_qzv"] = qzv

    scvi.settings.seed = 1
    fcm_orig = model.get_feature_correlation_matrix(
        correlation_type="pearson",
        n_samples=1,
        transform_batch=["batch_0", "batch_1"],
    )

    model.minify_adata()
    assert model.minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR

    scvi.settings.seed = 1
    fcm_new = model.get_feature_correlation_matrix(
        correlation_type="pearson",
        n_samples=1,
        transform_batch=["batch_0", "batch_1"],
    )

    assert_approx_equal(fcm_new, fcm_orig)



import sys
import warnings
from typing import Literal

import lightning.pytorch as pl
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.callbacks import LearningRateMonitor
from lightning.pytorch.loggers import Logger

from scvi import settings

from ._callbacks import (
    LoudEarlyStopping,
    SaveCheckpoint,
)
from ._logger import SimpleLogger
from ._progress import ProgressBar
from ._trainingplans import PyroTrainingPlan


class Trainer(pl.Trainer):
    """Lightweight wrapper of Pytorch Lightning Trainer.

    Appropriate defaults are set for scvi-tools models, as well as callbacks like
    EarlyStopping, with parameters accessible here.

    Parameters
    ----------
    accelerator
        Supports passing different accelerator types ("cpu", "gpu", "tpu", "ipu", "hpu", "mps,
        "auto") as well as custom accelerator instances.
    devices
        The devices to use. Can be set to a positive number (int or str), a sequence of device
        indices (list or str), the value ``-1`` to indicate all available devices should be used,
        or ``"auto"`` for automatic selection based on the chosen accelerator. Default: ``"auto"``.
    benchmark
        If true enables cudnn.benchmark, which improves speed when inputs are fixed size
    check_val_every_n_epoch
        Check val every n train epochs. By default, val is not checked, unless `early_stopping` is
        `True`.
    max_epochs
        Stop training once this number of epochs is reached.
    default_root_dir
        Default path for logs and weights when no logger/ckpt_callback passed.
        Defaults to `scvi.settings.logging_dir`. Can be remote file paths such as
        s3://mybucket/path or ‘hdfs://path/’
    enable_checkpointing
        If ``True``, enables checkpointing with a default :class:`~scvi.train.SaveCheckpoint`
        callback if there is no user-defined :class:`~scvi.train.SaveCheckpoint` in ``callbacks``.
    checkpointing_monitor
        If ``enable_checkpointing`` is ``True``, specifies the metric to monitor for checkpointing.
    num_sanity_val_steps
        Sanity check runs n validation batches before starting the training routine.
        Set it to -1 to run all batches in all validation dataloaders.
    enable_model_summary
        Whether to enable or disable the model summarization.
    early_stopping
        Whether to perform early stopping with respect to the validation set. This
        automatically adds a :class:`~lightning.pytorch.callbacks.early_stopping.EarlyStopping`
        instance. A custom instance can be passed by using the callbacks argument and
        setting this to `False`.
    early_stopping_monitor
        Metric logged during validation set epoch. The available metrics will depend on
        the training plan class used. We list the most common options here in the typing.
    early_stopping_min_delta
        Minimum change in the monitored quantity to qualify as an improvement,
        i.e. an absolute change of less than min_delta, will count as no improvement.
    early_stopping_patience
        Number of validation epochs with no improvement after which training will be stopped.
    early_stopping_mode
        In 'min' mode, training will stop when the quantity monitored has stopped decreasing
        and in 'max' mode it will stop when the quantity monitored has stopped increasing.
    enable_progress_bar
        Whether to enable or disable the progress bar.
    progress_bar_refresh_rate
        How often to refresh progress bar (in steps). Value 0 disables progress bar.
    simple_progress_bar
        Use custom scvi-tools simple progress bar (per epoch rather than per batch).
        When `False`, uses default PyTorch Lightning progress bar, unless `enable_progress_bar`
        is `False`.
    logger
        A valid pytorch lightning logger. Defaults to a simple dictionary logger.
        If `True`, defaults to the default pytorch lightning logger.
    log_every_n_steps
        How often to log within steps. This does not affect epoch-level logging.
    **kwargs
        Other keyword args for :class:`~pytorch_lightning.trainer.Trainer`
    """

    def __init__(
        self,
        accelerator: str | Accelerator | None = None,
        devices: list[int] | str | int | None = None,
        benchmark: bool = True,
        check_val_every_n_epoch: int | None = None,
        max_epochs: int = 400,
        default_root_dir: str | None = None,
        enable_checkpointing: bool = False,
        checkpointing_monitor: str = "validation_loss",
        num_sanity_val_steps: int = 0,
        enable_model_summary: bool = False,
        early_stopping: bool = False,
        early_stopping_monitor: Literal[
            "elbo_validation", "reconstruction_loss_validation", "kl_local_validation"
        ] = "elbo_validation",
        early_stopping_min_delta: float = 0.00,
        early_stopping_patience: int = 45,
        early_stopping_mode: Literal["min", "max"] = "min",
        enable_progress_bar: bool = True,
        progress_bar_refresh_rate: int = 1,
        simple_progress_bar: bool = True,
        logger: Logger | None | bool = None,
        log_every_n_steps: int = 10,
        learning_rate_monitor: bool = False,
        **kwargs,
    ):
        if default_root_dir is None:
            default_root_dir = settings.logging_dir

        check_val_every_n_epoch = check_val_every_n_epoch or sys.maxsize
        callbacks = kwargs.pop("callbacks", [])

        if early_stopping:
            early_stopping_callback = LoudEarlyStopping(
                monitor=early_stopping_monitor,
                min_delta=early_stopping_min_delta,
                patience=early_stopping_patience,
                mode=early_stopping_mode,
            )
            callbacks.append(early_stopping_callback)
            check_val_every_n_epoch = 1

        if enable_checkpointing and not any(isinstance(c, SaveCheckpoint) for c in callbacks):
            callbacks.append(SaveCheckpoint(monitor=checkpointing_monitor))
            check_val_every_n_epoch = 1
        elif any(isinstance(c, SaveCheckpoint) for c in callbacks):
            # check if user provided already provided the callback
            enable_checkpointing = True
            check_val_every_n_epoch = 1

        if learning_rate_monitor and not any(
            isinstance(c, LearningRateMonitor) for c in callbacks
        ):
            callbacks.append(LearningRateMonitor())
            check_val_every_n_epoch = 1

        if simple_progress_bar and enable_progress_bar:
            callbacks.append(ProgressBar(refresh_rate=progress_bar_refresh_rate))

        if logger is None:
            logger = SimpleLogger()

        super().__init__(
            accelerator=accelerator,
            devices=devices,
            benchmark=benchmark,
            check_val_every_n_epoch=check_val_every_n_epoch,
            max_epochs=max_epochs,
            default_root_dir=default_root_dir,
            enable_checkpointing=enable_checkpointing,
            num_sanity_val_steps=num_sanity_val_steps,
            enable_model_summary=enable_model_summary,
            logger=logger,
            log_every_n_steps=log_every_n_steps,
            enable_progress_bar=enable_progress_bar,
            callbacks=callbacks,
            **kwargs,
        )

    def fit(self, *args, **kwargs):
        """Fit the model."""
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action="ignore", category=UserWarning, message="The dataloader,"
            )
            warnings.filterwarnings(
                action="ignore",
                category=UserWarning,
                message="you defined a validation_step but have no val_dataloader",
            )
            warnings.filterwarnings(
                action="ignore",
                category=UserWarning,
                message="One of given dataloaders is None and it will be skipped",
            )
            # bug in pytorch lightning, assumes SequentialSampler
            # https://github.com/PyTorchLightning/pytorch-lightning/blob/
            # 48cb38ac5dd0159c8f7c5189c888dfd04a2ed34b/pytorch_lightning/
            # trainer/data_loading.py#L311-L314
            warnings.filterwarnings(
                action="ignore",
                category=UserWarning,
                message="Your `val_dataloader` has `shuffle=True`",
            )
            if isinstance(args[0], PyroTrainingPlan):
                warnings.filterwarnings(
                    action="ignore",
                    category=UserWarning,
                    message="`LightningModule.configure_optimizers` returned `None`",
                )
            super().fit(*args, **kwargs)



import numpy as np
import pytest

from scvi.data import synthetic_iid
from scvi.external import Decipher
from scvi.external.decipher.utils import Trajectory, rotate_decipher_components


@pytest.fixture(scope="session")
def adata():
    adata = synthetic_iid()
    return adata


def test_decipher_train(adata):
    Decipher.setup_anndata(adata)
    model = Decipher(adata)
    model.train(
        max_epochs=2,
        check_val_every_n_epoch=1,
        train_size=0.5,
        early_stopping=True,
    )


def test_decipher_model_methods(adata):
    Decipher.setup_anndata(adata)
    model = Decipher(adata)
    model.train(
        max_epochs=2,
        check_val_every_n_epoch=1,
        train_size=0.5,
        early_stopping=True,
    )
    v = model.get_latent_representation(give_z=False)
    z = model.get_latent_representation(give_z=True)
    assert v.shape == (adata.n_obs, model.module.dim_v)
    assert z.shape == (adata.n_obs, model.module.dim_z)

    v_obsm_key = "decipher_v"
    z_obsm_key = "decipher_z"
    adata.obsm[v_obsm_key] = v
    adata.obsm[z_obsm_key] = z

    v1_obs_col = "batch"
    rotated_v, rotated_z, rotation = rotate_decipher_components(
        adata,
        v_obsm_key=v_obsm_key,
        z_obsm_key=z_obsm_key,
        v1_obs_col=v1_obs_col,
        v1_order=["batch_0", "batch_1"],
        auto_flip_decipher_z=True,
    )
    assert rotated_v.shape == (adata.n_obs, model.module.dim_v)
    assert not np.allclose(rotated_v, v)
    assert rotated_z.shape == (adata.n_obs, model.module.dim_z)
    assert not np.allclose(rotated_z, z)
    assert rotation.shape == (model.module.dim_v, model.module.dim_v)

    imputed_expr, v_cov, z_cov = model.compute_imputed_gene_expression(
        adata,
        compute_covariances=True,
        v_obsm_key=v_obsm_key,
        z_obsm_key=z_obsm_key,
    )
    assert imputed_expr.shape == (adata.n_obs, adata.n_vars)
    assert v_cov.shape == (adata.n_vars, model.module.dim_v)
    assert z_cov.shape == (adata.n_vars, model.module.dim_z)


def test_decipher_trajectory(adata):
    Decipher.setup_anndata(adata)
    model = Decipher(adata)
    model.train(
        max_epochs=2,
        check_val_every_n_epoch=1,
        train_size=0.5,
        early_stopping=True,
    )

    v = model.get_latent_representation(give_z=False)
    adata.obsm["decipher_v"] = v

    trajectory = Trajectory.from_dict(
        {
            "rep_key": "decipher_v",
            "cluster_locations": v[
                np.random.choice(np.arange(v.shape[0]), size=10, replace=False), :
            ],
            "cluster_ids": np.arange(10),
            "density": 50,
        }
    )

    adata.obs["cluster_ids"] = np.random.randint(0, 10, size=adata.n_obs)
    decipher_time = model.compute_decipher_time(
        adata,
        cluster_obs_key="cluster_ids",
        trajectory=trajectory,
        n_neighbors=10,
    )
    assert decipher_time.shape == (adata.n_obs,)

    gene_patterns = model.compute_gene_patterns(
        adata,
        trajectory=trajectory,
        l_scale=10_000,
        n_samples=100,
    )
    assert gene_patterns["mean"].shape[1] == adata.n_vars
    assert gene_patterns["q25"].shape[1] == adata.n_vars
    assert gene_patterns["q75"].shape[1] == adata.n_vars
    n_times = gene_patterns["mean"].shape[0]
    assert gene_patterns["times"].shape == (n_times,)



from __future__ import annotations

import logging
import os
import warnings
from typing import TYPE_CHECKING

import anndata
import mudata
import numpy as np
import pandas as pd
import torch
from anndata import AnnData, read_h5ad

from scvi import settings
from scvi.data._constants import _SETUP_METHOD_NAME
from scvi.data._download import _download
from scvi.model.base._constants import SAVE_KEYS

if TYPE_CHECKING:
    from typing import Literal

    import numpy.typing as npt

    from scvi._types import AnnOrMuData

logger = logging.getLogger(__name__)


def _load_legacy_saved_files(
    dir_path: str,
    file_name_prefix: str,
    load_adata: bool,
) -> tuple[dict, np.ndarray, dict, AnnData | None]:
    model_path = os.path.join(dir_path, f"{file_name_prefix}{SAVE_KEYS.LEGACY_MODEL_FNAME}")
    var_names_path = os.path.join(
        dir_path, f"{file_name_prefix}{SAVE_KEYS.LEGACY_VAR_NAMES_FNAME}"
    )
    setup_dict_path = os.path.join(
        dir_path, f"{file_name_prefix}{SAVE_KEYS.LEGACY_SETUP_DICT_FNAME}"
    )

    model_state_dict = torch.load(model_path, map_location="cpu", weights_only=False)

    var_names = np.genfromtxt(var_names_path, delimiter=",", dtype=str)

    with open(setup_dict_path, "rb") as handle:
        attr_dict = pd.read_pickle(handle)

    if load_adata:
        adata_path = os.path.join(dir_path, f"{file_name_prefix}{SAVE_KEYS.ADATA_FNAME}")
        if os.path.exists(adata_path):
            adata = read_h5ad(adata_path)
        elif not os.path.exists(adata_path):
            raise ValueError("Save path contains no saved anndata and no adata was passed.")
    else:
        adata = None

    return model_state_dict, var_names, attr_dict, adata


def _load_saved_files(
    dir_path: str,
    load_adata: bool,
    prefix: str | None = None,
    map_location: Literal["cpu", "cuda"] | None = None,
    backup_url: str | None = None,
) -> tuple[dict, np.ndarray, dict, AnnData]:
    """Helper to load saved files."""
    file_name_prefix = prefix or ""

    model_file_name = f"{file_name_prefix}{SAVE_KEYS.MODEL_FNAME}"
    model_path = os.path.join(dir_path, model_file_name)
    try:
        _download(backup_url, dir_path, model_file_name)
        model = torch.load(model_path, map_location=map_location, weights_only=False)
    except FileNotFoundError as exc:
        raise ValueError(
            f"Failed to load model file at {model_path}. "
            "If attempting to load a saved model from <v0.15.0, please use the util function "
            "`convert_legacy_save` to convert to an updated format."
        ) from exc

    model_state_dict = model.get(SAVE_KEYS.MODEL_STATE_DICT_KEY)
    var_names = model.get(SAVE_KEYS.VAR_NAMES_KEY)
    attr_dict = model.get(SAVE_KEYS.ATTR_DICT_KEY)

    if load_adata:
        is_mudata = attr_dict["registry_"].get(_SETUP_METHOD_NAME) == "setup_mudata"
        file_suffix = SAVE_KEYS.ADATA_FNAME if is_mudata is False else SAVE_KEYS.MDATA_FNAME
        adata_path = os.path.join(dir_path, f"{file_name_prefix}{file_suffix}")
        if os.path.exists(adata_path):
            if is_mudata:
                adata = mudata.read(adata_path)
            else:
                adata = anndata.read_h5ad(adata_path)
        else:
            raise ValueError("Save path contains no saved anndata and no adata was passed.")
    else:
        adata = None

    return attr_dict, var_names, model_state_dict, adata


def _initialize_model(cls, adata, attr_dict):
    """Helper to initialize a model."""
    if "init_params_" not in attr_dict.keys():
        raise ValueError(
            "No init_params_ were saved by the model. Check out the "
            "developers guide if creating custom models."
        )
    # get the parameters for the class init signature
    init_params = attr_dict.pop("init_params_")

    # new saving and loading, enable backwards compatibility
    if "non_kwargs" in init_params.keys():
        # grab all the parameters except for kwargs (is a dict)
        non_kwargs = init_params["non_kwargs"]
        kwargs = init_params["kwargs"]

        # expand out kwargs
        kwargs = {k: v for (i, j) in kwargs.items() for (k, v) in j.items()}
    else:
        # grab all the parameters except for kwargs (is a dict)
        non_kwargs = {k: v for k, v in init_params.items() if not isinstance(v, dict)}
        kwargs = {k: v for k, v in init_params.items() if isinstance(v, dict)}
        kwargs = {k: v for (i, j) in kwargs.items() for (k, v) in j.items()}
        non_kwargs.pop("use_cuda")

    # backwards compat for scANVI
    if "unlabeled_category" in non_kwargs.keys():
        non_kwargs.pop("unlabeled_category")
    if "pretrained_model" in non_kwargs.keys():
        non_kwargs.pop("pretrained_model")

    model = cls(adata, **non_kwargs, **kwargs)
    for attr, val in attr_dict.items():
        setattr(model, attr, val)

    return model


def _get_var_names(
    adata: AnnOrMuData,
    legacy_mudata_format: bool = False,
) -> npt.NDArray | dict[str, npt.NDArray]:
    """Get variable names from an :class:`~anndata.AnnData` or :class:`~mudata.MuData` object.

    Parameters
    ----------
    adata
        :class:`~anndata.AnnData` or :class:`~mudata.MuData` object.
    legacy_mudata_format
        If ``True``, returns variable names for :class:`~mudata.MuData` objects in the legacy
        format, _i.e._, a flat array with variable names across all modalities concatenated
        together. Otherwise, returns a dictionary with keys corresponding to the modality names
        and values corresponding to the variable names for each modality.

    Returns
    -------
    - An array of variable names if ``adata`` is an :class:`~anndata.AnnData` object
    - An array of concatenated modality variable names if ``adata`` is a :class:`~mudata.MuData`
        and ``legacy_mudata_format`` is ``True``
    - A dictionary with keys corresponding to the modality names and values corresponding to the
        variable names for each modality if ``adata`` is a :class:`~mudata.MuData` and
        ``legacy_mudata_format`` is ``False``
    """
    if isinstance(adata, AnnData) or legacy_mudata_format:
        return adata.var_names.astype(str).to_numpy()
    elif isinstance(adata, mudata.MuData):
        return {
            mod_key: adata.mod[mod_key].var_names.astype(str).to_numpy()
            for mod_key in adata.mod.keys()
        }
    else:
        raise TypeError("`adata` must be an AnnData or MuData object.")


def _validate_var_names(
    adata: AnnOrMuData, source_var_names: npt.NDArray | dict[str, npt.NDArray]
) -> None:
    """Validate that source and loaded variable names match.

    Parameters
    ----------
    adata
        :class:`~anndata.AnnData` or :class:`~mudata.MuData` object.
    source_var_names
        Variable names from a saved model file corresponding to the variable names used during
        training.
    """
    from numpy import array_equal

    is_anndata = isinstance(adata, AnnData)
    source_per_mod_var_names = isinstance(source_var_names, dict)
    load_var_names = _get_var_names(
        adata,
        legacy_mudata_format=(not is_anndata and not source_per_mod_var_names),
    )

    if source_per_mod_var_names:
        valid_load_var_names = all(
            array_equal(source_var_names.get(mod_key), load_var_names.get(mod_key))
            for mod_key in source_var_names
        )
    else:
        valid_load_var_names = array_equal(source_var_names, load_var_names)

    if not valid_load_var_names:
        warnings.warn(
            "`var_names` for the loaded `adata` does not match those of the `adata` used to "
            "train the model. For valid results, the former should match the latter.",
            UserWarning,
            stacklevel=settings.warnings_stacklevel,
        )



import os
import random

import anndata
import numpy as np
import pandas as pd
import pytest
import torch
from scipy.sparse import csr_matrix

import scvi
from scvi import REGISTRY_KEYS
from scvi.data import AnnTorchDataset, _constants, synthetic_iid
from scvi.data.fields import ObsmField, ProteinObsmField

from .utils import generic_setup_adata_manager


def test_transfer_fields_basic(adata1, adata2):
    # test transfer_fields function
    adata2.X = adata1.X
    adata1_manager = generic_setup_adata_manager(adata1)
    adata1_manager.transfer_fields(adata2)
    np.testing.assert_array_equal(adata1.obs["_scvi_labels"], adata2.obs["_scvi_labels"])


def test_transfer_fields_layer_use(adata1, adata2):
    # test if layer was used initially, again used in transfer setup
    zeros = np.zeros_like(adata1.X)
    ones = np.ones_like(adata1.X)
    adata1.X = zeros
    adata2.X = ones
    adata1_manager = generic_setup_adata_manager(adata1, layer="raw")
    adata1_manager.transfer_fields(adata2)
    np.testing.assert_array_equal(adata1.obs["_scvi_labels"], adata2.obs["_scvi_labels"])


def test_transfer_fields_unknown_batch(adata1, adata2):
    # test that an unknown batch throws an error
    adata2.obs["batch"] = [2] * adata2.n_obs
    adata1_manager = generic_setup_adata_manager(adata1, batch_key="batch")
    with pytest.raises(ValueError):
        adata1_manager.transfer_fields(adata2)


def test_transfer_fields_unknown_label(adata1, adata2):
    # test that an unknown label throws an error
    adata2.obs["labels"] = ["label_123"] * adata2.n_obs
    adata1_manager = generic_setup_adata_manager(adata1, labels_key="labels")
    with pytest.raises(ValueError):
        adata1_manager.transfer_fields(adata2)


def test_transfer_fields_correct_mapping(adata1, adata2):
    # test that correct mapping was applied
    adata2.obs["labels"] = ["label_1"] * adata2.n_obs
    adata1_manager = generic_setup_adata_manager(adata1, labels_key="labels")
    adata1_manager.transfer_fields(adata2)
    labels_mapping = adata1_manager.get_state_registry("labels").categorical_mapping
    correct_label = np.where(labels_mapping == "label_1")[0][0]
    assert adata2.obs._scvi_labels.iloc[0] == correct_label


def test_transfer_fields_correct_batch(adata1, adata2):
    # test that transfer_fields correctly looks for adata.obs['batch']
    del adata2.obs["batch"]
    adata1_manager = generic_setup_adata_manager(adata1, batch_key="batch")
    with pytest.raises(KeyError):
        adata1_manager.transfer_fields(adata2)


def test_transfer_fields_same_batch_and_label(adata1, adata2):
    # test that transfer_fields assigns same batch and label to cells
    # if the original anndata was also same batch and label
    adata1_manager = generic_setup_adata_manager(adata1)
    del adata2.obs["batch"]
    adata1_manager.transfer_fields(adata2)
    assert adata2.obs._scvi_batch.iloc[0] == 0
    assert adata2.obs._scvi_labels.iloc[0] == 0


def test_transfer_fields_subset(adata1, adata2):
    # test that if a category mapping is a subset, transfer anndata is called
    scvi.model.SCVI.setup_anndata(adata1, batch_key="batch")
    adata2.obs["batch"] = "batch_1"
    scvi.model.SCVI.setup_anndata(adata2, batch_key="batch")
    m = scvi.model.SCVI(adata1)
    m.train(1)
    m.get_latent_representation(adata2)
    assert adata2.obs["_scvi_batch"].all() == 1


def test_transfer_fields_wrong_kwarg(adata):
    # test that error is thrown if an arbitrary kwarg is passed into setup_anndata
    with pytest.raises(TypeError):
        scvi.model.SCVI.setup_anndata(adata, batch="batch")


def test_clobber_same_model(adata):
    scvi.model.SCVI.setup_anndata(adata)
    m1 = scvi.model.SCVI(adata)
    m1.train(1)

    scvi.model.SCVI.setup_anndata(adata, batch_key="batch")
    m2 = scvi.model.SCVI(adata)
    m2.train(1)

    adata_manager1 = m1.get_anndata_manager(adata)
    assert adata_manager1.summary_stats.n_batch == 1
    # The underlying data is still 2 since we have not run _validate_anndata yet
    # to re-transfer the setup of m1.
    assert len(np.unique(adata_manager1.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 2
    m1._validate_anndata(adata)
    assert len(np.unique(adata_manager1.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 1

    adata_manager2 = m2.get_anndata_manager(adata)
    assert adata_manager2.summary_stats.n_batch == 2
    # The underlying data is still 1 since we have not run _validate_anndata yet
    # to re-transfer the setup of m2.
    assert len(np.unique(adata_manager2.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 1
    m2._validate_anndata(adata)
    assert len(np.unique(adata_manager2.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 2


def test_clobber_different_models(adata):
    scvi.model.SCVI.setup_anndata(adata, batch_key="batch")
    m1 = scvi.model.SCVI(adata)
    m1.train(1)

    scvi.model.TOTALVI.setup_anndata(
        adata,
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    m2 = scvi.model.TOTALVI(adata)
    m2.train(1)

    adata_manager1 = m1.get_anndata_manager(adata)
    assert adata_manager1.summary_stats.n_batch == 2
    # The underlying data is still 2 since we have not run _validate_anndata yet
    # to re-transfer the setup of m1.
    assert len(np.unique(adata_manager1.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 1
    m1._validate_anndata(adata)
    assert len(np.unique(adata_manager1.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 2

    adata_manager2 = m2.get_anndata_manager(adata)
    assert adata_manager2.summary_stats.n_batch == 1
    # The underlying data is still 1 since we have not run _validate_anndata yet
    # to re-transfer the setup of m2.
    assert len(np.unique(adata_manager2.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 2
    m2._validate_anndata(adata)
    assert len(np.unique(adata_manager2.get_from_registry(REGISTRY_KEYS.BATCH_KEY))) == 1


def test_register_new_fields(adata):
    bdata = adata.copy()
    incremental_adata_manager = generic_setup_adata_manager(bdata)
    batch_field = None
    for field in incremental_adata_manager.fields:
        if field.registry_key == REGISTRY_KEYS.BATCH_KEY:
            batch_field = field
    new_fields = [
        ProteinObsmField(
            REGISTRY_KEYS.PROTEIN_EXP_KEY,
            "protein_expression",
            batch_field=batch_field,
            use_batch_mask=True,
            is_count_data=True,
        )
    ]
    incremental_adata_manager.register_new_fields(new_fields)
    adata_manager = generic_setup_adata_manager(
        adata, protein_expression_obsm_key="protein_expression"
    )
    np.testing.assert_array_equal(
        incremental_adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
    )
    assert len(incremental_adata_manager.fields) == len(adata_manager.fields)


def test_register_new_fields_with_transferred_manager(adata):
    bdata = adata.copy()
    cdata = adata.copy()
    adata_manager = generic_setup_adata_manager(adata, batch_key="batch")
    bdata.obs["batch"] = adata.obs["batch"].to_numpy()[0]
    bdata_manager = adata_manager.transfer_fields(bdata)
    new_fields = [ObsmField(REGISTRY_KEYS.PROTEIN_EXP_KEY, "protein_expression")]
    bdata_manager.register_new_fields(new_fields)
    cdata_manager = bdata_manager.transfer_fields(cdata)

    # Should have protein field
    cdata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY)
    np.testing.assert_array_equal(cdata.obs["_scvi_batch"].values, adata.obs["_scvi_batch"].values)


def test_update_setup_args(adata):
    adata_manager = generic_setup_adata_manager(adata)
    adata_manager.update_setup_method_args({"test_arg": "test_val"})
    assert "test_arg" in adata_manager._get_setup_method_args()[_constants._SETUP_ARGS_KEY].keys()


def test_data_format(adata):
    # if data was dense np array, check after setup_anndata, data is C_CONTIGUOUS
    old_x = adata.X
    old_pro = adata.obsm["protein_expression"]
    old_obs = adata.obs
    adata.X = np.asfortranarray(old_x)
    adata.obsm["protein_expression"] = np.asfortranarray(old_pro)
    assert adata.X.flags["C_CONTIGUOUS"] is False
    assert adata.obsm["protein_expression"].flags["C_CONTIGUOUS"] is False

    adata_manager = generic_setup_adata_manager(
        adata, protein_expression_obsm_key="protein_expression"
    )
    assert adata.X.flags["C_CONTIGUOUS"] is True
    assert adata.obsm["protein_expression"].flags["C_CONTIGUOUS"] is True

    assert np.array_equal(old_x, adata.X)
    assert np.array_equal(old_pro, adata.obsm["protein_expression"])
    assert np.array_equal(old_obs, adata.obs)

    assert np.array_equal(adata.X, adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY))
    assert np.array_equal(
        adata.obsm["protein_expression"],
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
    )


def test_data_format_c_contiguous(adata):
    # if obsm is dataframe, make it C_CONTIGUOUS if it isnt
    pe = np.asfortranarray(adata.obsm["protein_expression"])
    adata.obsm["protein_expression"] = pd.DataFrame(pe, index=adata.obs_names)
    assert adata.obsm["protein_expression"].to_numpy().flags["C_CONTIGUOUS"] is False
    adata_manager = generic_setup_adata_manager(
        adata, protein_expression_obsm_key="protein_expression"
    )
    new_pe = adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY)
    assert new_pe.to_numpy().flags["C_CONTIGUOUS"] is True
    assert np.array_equal(pe, new_pe)
    assert np.array_equal(adata.X, adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY))
    assert np.array_equal(
        adata.obsm["protein_expression"],
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
    )


def test_setup_anndata(adata):
    # test regular setup
    adata_manager = generic_setup_adata_manager(
        adata,
        batch_key="batch",
        labels_key="labels",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY),
        np.array(adata.obs["_scvi_batch"]).reshape((-1, 1)),
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.LABELS_KEY),
        np.array(adata.obs["labels"].cat.codes).reshape((-1, 1)),
    )
    np.testing.assert_array_equal(adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY), adata.X)
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY),
        adata.obsm["protein_expression"],
    )
    np.testing.assert_array_equal(
        adata_manager.get_state_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY).column_names,
        adata.uns["protein_names"],
    )


def test_setup_anndata_view_error(adata):
    # test that error is thrown if its a view:
    with pytest.raises(ValueError):
        generic_setup_adata_manager(adata[1])


def test_setup_anndata_view_error_df_protein_none(adata):
    # If obsm is a df and protein_names_uns_key is None, protein names should be grabbed from
    # column of df
    new_protein_names = np.array(random.sample(range(100), 100)).astype("str")
    df = pd.DataFrame(
        adata.obsm["protein_expression"],
        index=adata.obs_names,
        columns=new_protein_names,
    )
    adata.obsm["protein_expression"] = df
    adata_manager = generic_setup_adata_manager(
        adata, protein_expression_obsm_key="protein_expression"
    )
    np.testing.assert_array_equal(
        adata_manager.get_state_registry(REGISTRY_KEYS.PROTEIN_EXP_KEY).column_names,
        new_protein_names,
    )


def test_setup_anndata_layer(adata):
    # test that layer is working properly
    true_x = adata.X
    adata.layers["X"] = true_x
    adata.X = np.ones_like(adata.X)
    adata_manager = generic_setup_adata_manager(adata, layer="X")
    np.testing.assert_array_equal(adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY), true_x)


def test_setup_anndata_create_label_batch(adata):
    # test that it creates labels and batch if no layers_key is passed
    adata_manager = generic_setup_adata_manager(
        adata,
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.BATCH_KEY),
        np.zeros((adata.shape[0], 1)),
    )
    np.testing.assert_array_equal(
        adata_manager.get_from_registry(REGISTRY_KEYS.LABELS_KEY),
        np.zeros((adata.shape[0], 1)),
    )


def test_setup_anndata_nan(adata):
    # test error is thrown when categorical obs field contains nans
    adata.obs["batch"][:10] = np.nan
    with pytest.raises(ValueError):
        generic_setup_adata_manager(adata, batch_key="batch")


def test_setup_anndata_cat(adata):
    # test error is thrown when categorical joint obsm field contains nans
    adata.obs["cat1"][:10] = np.nan
    with pytest.raises(ValueError):
        generic_setup_adata_manager(adata, categorical_covariate_keys=["cat1"])


def test_save_setup_anndata(adata, save_path):
    generic_setup_adata_manager(
        adata,
        batch_key="batch",
        labels_key="labels",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    adata.write(os.path.join(save_path, "test.h5ad"))


def test_extra_covariates(adata):
    scvi.model.SCVI.setup_anndata(
        adata,
        batch_key="batch",
        labels_key="labels",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    m = scvi.model.SCVI(adata)
    m.train(1)
    df1 = m.get_from_registry(adata, REGISTRY_KEYS.CONT_COVS_KEY)
    df2 = adata.obs[["cont1", "cont2"]]
    pd.testing.assert_frame_equal(df1, df2)


def test_extra_covariates_transfer(adata):
    adata_manager = generic_setup_adata_manager(
        adata,
        batch_key="batch",
        labels_key="labels",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    bdata = synthetic_iid()
    bdata.obs["cont1"] = np.random.normal(size=(bdata.shape[0],))
    bdata.obs["cont2"] = np.random.normal(size=(bdata.shape[0],))
    bdata.obs["cat1"] = 0
    bdata.obs["cat2"] = 1

    adata_manager.transfer_fields(bdata)

    # give it a new category
    bdata.obs["cat1"] = 6
    bdata_manager = adata_manager.transfer_fields(bdata, extend_categories=True)
    assert bdata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY).mappings["cat1"][-1] == 6


def test_anntorchdataset_getitem(adata):
    adata_manager = generic_setup_adata_manager(
        adata,
        batch_key="batch",
        labels_key="labels",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
    )
    # check that we can successfully pass in a list of tensors to get
    tensors_to_get = [REGISTRY_KEYS.BATCH_KEY, REGISTRY_KEYS.LABELS_KEY]
    bd = AnnTorchDataset(adata_manager, getitem_tensors=tensors_to_get)
    np.testing.assert_array_equal(tensors_to_get, list(bd[1].keys()))

    # check that we can successfully pass in a dict of tensors and their associated types
    bd = AnnTorchDataset(
        adata_manager,
        getitem_tensors={
            REGISTRY_KEYS.X_KEY: np.int64,
            REGISTRY_KEYS.LABELS_KEY: np.int64,
        },
    )
    assert bd[1][REGISTRY_KEYS.X_KEY].dtype == np.int64
    assert bd[1][REGISTRY_KEYS.LABELS_KEY].dtype == np.int64

    # check that by default we get all the registered tensors
    bd = AnnTorchDataset(adata_manager)
    all_registered_tensors = list(adata_manager.data_registry.keys())
    np.testing.assert_array_equal(all_registered_tensors, list(bd[1].keys()))
    assert bd[1][REGISTRY_KEYS.X_KEY].shape == (
        1,
        bd.adata_manager.summary_stats.n_vars,
    )


def test_anntorchdataset_from_manager(adata):
    # check that AnnTorchDataset returns numpy array
    adata_manager = generic_setup_adata_manager(adata)
    bd = adata_manager.create_torch_dataset()
    assert isinstance(bd, AnnTorchDataset)

    bd = adata_manager.create_torch_dataset(indices=np.arange(adata.n_obs))
    assert isinstance(bd, torch.utils.data.Subset)


def test_anntorchdataset_numpy(adata):
    # check that AnnTorchDataset returns numpy array
    adata_manager = generic_setup_adata_manager(adata)
    bd = AnnTorchDataset(adata_manager)
    for value in bd[1].values():
        assert isinstance(value, np.ndarray)


def test_anntorchdataset_numpy_sparse(adata):
    # check AnnTorchDataset returns numpy array counts were sparse
    adata.X = csr_matrix(adata.X)
    adata_manager = generic_setup_adata_manager(adata)
    bd = AnnTorchDataset(adata_manager)
    for value in bd[1].values():
        assert isinstance(value, np.ndarray)


def test_anntorchdataset_getitem_numpy_sparse(adata):
    # check AnnTorchDataset returns numpy array if pro exp was sparse
    adata.obsm["protein_expression"] = csr_matrix(adata.obsm["protein_expression"])
    adata_manager = generic_setup_adata_manager(
        adata, batch_key="batch", protein_expression_obsm_key="protein_expression"
    )
    bd = AnnTorchDataset(adata_manager)
    for value in bd[1].values():
        assert isinstance(value, np.ndarray)


def test_anntorchdataset_getitem_pro_exp(adata):
    # check pro exp is being returned as numpy array even if its DF
    adata.obsm["protein_expression"] = pd.DataFrame(
        adata.obsm["protein_expression"], index=adata.obs_names
    )
    adata_manager = generic_setup_adata_manager(
        adata, batch_key="batch", protein_expression_obsm_key="protein_expression"
    )
    bd = AnnTorchDataset(adata_manager)
    for value in bd[1].values():
        assert isinstance(value, np.ndarray)


def test_view_registry(adata):
    adata_manager = generic_setup_adata_manager(
        adata,
        batch_key="batch",
        labels_key="labels",
        protein_expression_obsm_key="protein_expression",
        protein_names_uns_key="protein_names",
        continuous_covariate_keys=["cont1", "cont2"],
        categorical_covariate_keys=["cat1", "cat2"],
    )
    adata_manager.view_registry()
    adata_manager.view_registry(hide_state_registries=True)


def test_saving(adata, save_path):
    save_path = os.path.join(save_path, "tmp_adata.h5ad")
    adata.obs["cat1"] = np.random.randint(0, 3, adata.n_obs).astype(str)
    adata.obs["cat1"][1] = "asdf"
    adata.obs["cat1"][2] = "f34"
    adata.obs["cat2"] = np.random.randint(0, 7, adata.n_obs)

    generic_setup_adata_manager(
        adata,
        protein_expression_obsm_key="protein_expression",
        batch_key="batch",
        labels_key="labels",
        categorical_covariate_keys=["cat1", "cat2"],
        continuous_covariate_keys=["cont1", "cont2"],
    )
    adata.write(save_path)
    anndata.read_h5ad(save_path)


def test_backed_anndata(adata, save_path):
    path = os.path.join(save_path, "test_data.h5ad")
    adata.write_h5ad(path)
    adata = anndata.read_h5ad(path, backed="r+")
    adata_manager = generic_setup_adata_manager(adata, batch_key="batch")

    # test get item
    bd = AnnTorchDataset(adata_manager)
    subset = bd[np.arange(adata.n_obs)]
    assert isinstance(subset["X"], np.ndarray)


def test_backed_anndata_sparse(adata, save_path):
    # sparse
    adata.X = csr_matrix(adata.X)
    path = os.path.join(save_path, "test_data2.h5ad")
    adata.write_h5ad(path)
    adata = anndata.read_h5ad(path, backed="r+")
    adata_manager = generic_setup_adata_manager(adata, batch_key="batch")

    # test get item
    bd = AnnTorchDataset(adata_manager)
    subset = bd[np.arange(adata.n_obs)]
    assert isinstance(subset["X"], np.ndarray)
